#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <algorithm>

#include "Solver.h"
#include "SimpSolver.h"
using namespace std;

#include "d0efs.hh"



/*================================================

  R E A D _ D I M A C S 1

  Returns 0 if formula's format checked out ok.
  In the case of a "minor" problem returns 1.
  In the  case of "severe" problems the "exit"
  function is called

  ===============================================*/
int read_dimacs1(FILE *fp,DNF &D,int &num_vars) 
{
  int nvars,ncubes,c;
  char buf[BUF_SIZE];
  CUBE C;
  int ret_value=0;
  /* -------------------------------
     Read in the  number of variables
     and cubes skipping comments
     --------------------------------*/
  while (true) {
    if (fgets(buf,BUF_SIZE,fp) == (char *) NULL) {
      cout << "eof is encountered\n";
      exit(1);
    }
      
    switch (buf[0]) {
    case 'c': /* comment */
      continue; /* read in next line */
    case 'p':  /* the first informative line */
      c = sscanf(buf,"p cnf %d %d", &nvars, &ncubes);
      assert((nvars >= 0) && (ncubes >= 0));
      if ((c == EOF) || (c != 2)) {
        cout << "wrong format " << buf << endl;
        exit(1);
      }
      goto START;
    default:
      {cout << "wrong format " << buf <<  endl;
       exit(1);
      } 
    }
  }
  /* --------------------------
     Read in the body
     -------------------------*/
 START:  
  SCUBE S; 
  int count=1;
  int empty_cube=0;

  while (c != EOF) {
    int v;
    
 
    c = fscanf(fp,"%d", &v);
    if (c == 1) {
      if (v!=0) {
        if (S.find(-1*v)!=S.end()) { // in the current cube we have already seen the opposite literal
          empty_cube = 1;
          ret_value = 1;
          fprintf(stderr,"cube number %d is empty!\n",count);
          goto read_next;              
        }
           	
        if (S.find(v)!=S.end()) { // in the current cube we have already seen the same literal
          ret_value = 1;
          goto read_next; // just skip adding the literal to the current cube
        }
        // none of the two alternatives above has happened
        // so we add literal v to s and add v to the current cube  
        S.insert(v);
        C.push_back(v);
           
        if (abs(v) > nvars) {
          fprintf(stderr,"variable number (%d) exceeds the declared number of variables (%d)\n",v,nvars);
          exit(1);
        }
      } else { // symbol 0 (end of cube) is read in
        //if (!cube.empty() && !empty_cube) 
        if (!empty_cube) // we allow cube-universe but ignore cubes with contradictory literals
          D.push_back(C);
        count++;
        C.clear();
        S.clear();
        empty_cube = 0;
      }
              
    }	 
  read_next:;
  }/* end of the while loop */

  fclose(fp);
  if (D.size() > unsigned(ncubes))   {
    fprintf(stderr,"the seen number of cubes (%d) exceeds the declared number of cubes (%d)\n",
            (int) D.size(),ncubes);
    exit(1);
  }

  num_vars = nvars;

  return (ret_value);  
}/* end of function read_dimacs1 */


/*==============================

  F I N D _ M A X _ V A R

  ===============================*/
int find_max_var(DNF &D)

{
  int max=0;

  for (uint i = 0; i < D.size();i++) {
    CUBE &C = D[i];
    for (uint j = 0; j < C.size();j++)
      if (abs(C[j]) >  max)
        max = abs(C[j]);
  }
  
  return(max);

} /* end of function find_max_var */
