#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>


#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;

#include "d0efs.hh"


/*===================

      M A I N

  ==================*/
int main(int argc,char *argv[])
{

  EgPqe  Ep;

  Ep.time_limit = -1;
  Ep.max_new_ycubes = -1;
  
  if (argc == 1)  {
    printf("eg-pqe root ['e'num] ['t'val]\n");
    printf("'e'num   - terminate if the solution contains 'num' clauses or more \n");
    printf("'t'val   - sets the time limit (in seconds)\n");
    exit(0);
  }

  if (argc > 4)   {
    printf("wrong number of arguments: %d\n",argc);
    exit(ERROR); }

  Ep.read_input(argv[1]);
  for (int i=2; i < argc; i++) {
    switch (argv[i][0])     {
    case 'e':
      Ep.max_new_ycubes = atoi(argv[i]+1);
      break; 
    case 't':
      Ep.time_limit = atof(argv[i]+1);
      if (Ep.time_limit <= 0)   {
        printf("wrong time limit %s\n",argv[i]);
        exit(ERROR);
      }
      break;
    default:
      printf("wrong parameter %s\n",argv[i]);
      exit(100);
    }

  }

  if (Ep.time_limit > 0) 
    Minisat::limitTime(Ep.time_limit+10);
  
  double usrtime0=0.,systime0=0.;
  double usrtime=0.,systime=0.;
  get_runtime (usrtime0, systime0);
  
  Ep.init();
  bool fin = Ep.eg_pqe();
  Ep.print_stat();
  get_runtime (usrtime, systime);
  printf("total time is %.2f sec.\n",usrtime-usrtime0);
  char Fname[MAX_NAME];
  strcpy(Fname,argv[1]);
  strcat(Fname,".res.cnf");
  print_dnf(Ep.Sol,Fname,Ep.num_vars);
  if (fin) {
    printf("successful termination\n");   
    exit(0);
  }
  else {
    printf("not finished\n");
    exit(1);
  }

}/* end of function main */


//
//  EG_PQE
//
bool EgPqe::eg_pqe()
{

  for (uint i=0; i < Sol.size(); i++)
    if (Sol[i].size() == 0) return(true);
  
  while (F1.size() > 0) {
    CLAUSE C = F1.back();
    F1.pop_back();
    bool ok = prove_red(C);
    if (!ok) return(false);
  }

  return(true);

} /* eg_pqe */


//
//  READ_INPUT
//
void EgPqe::read_input(char *root)
{  
  char fname[MAX_NAME];
  strcpy(fname,root);

  // read CNF formula
  strcat(fname,".cnf");
  FILE *fp1 = fopen(fname,"r");
  if (fp1 == NULL) {
    p();
    exit(ERROR);
  }
  read_dimacs1(fp1,F,num_vars);
  
  // read list of free (i.e. unquantified) variables
  fname[0] = 0;
  strcpy(fname,root);
  strcat(fname,".txt");
  read_numbers(fname,Yvars);   
 
  
  // read list of cubes to be made redundant
  fname[0] = 0;
  strcpy(fname,root);
  strcat(fname,".clauses");
  read_numbers(fname,Trgs);
  
}  /* read_input */


//
//   INIT
//
void EgPqe::init()
{

  
  Xvar.assign(num_vars,1);
  for (uint i=0; i < Yvars.size(); i++)
    Xvar[Yvars[i]-1] = 0;

  SCUBE S;
  for (uint i=0; i < Trgs.size(); i++)
    S.insert(Trgs[i]);

  for (uint i=0; i < F.size(); i++)
    if (S.find(i) != S.end()) F1.push_back(F[i]);
    else F2.push_back(F[i]);


  for (int i=0; i < num_vars; i++)    {
    CUBE  Dummy;
    Flits0.push_back(Dummy);
    Flits1.push_back(Dummy);
  }

  // initialize literal counts  
  initialize_cube_vects(F,Flits0,Flits1);

  Slvr = init_sat_solver(num_vars);
  accept_new_clauses(Slvr,F);

  new_ycubes = 0;
  rem_bnd_pnts = 0;
  unrem_bnd_pnts = 0;
} /* init */

//
//  PRINT_STAT
//
void EgPqe::print_stat()
{
  printf("#removable boundary points %d\n",rem_bnd_pnts);
  printf("#unremovable boundary points %d\n",unrem_bnd_pnts);
         

} /* print_stat */
