#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
using namespace std;

#include "d0efs.hh"

//
//  PROVE_RED
//
bool EgPqe::prove_red(CLAUSE &C)
{

  Msolver *Aux_slvr = init_sat_solver(num_vars);

  accept_new_clauses(Aux_slvr,F1);
  accept_new_clauses(Aux_slvr,F2);
  accept_new_clauses(Aux_slvr,Sol);

  while (true) {
    bool ok = check_time();
    if (!ok) return(false);
    MvecLits Assmps;
    add_negated_assumps(Assmps,C);
    bool sat = Aux_slvr->solve(Assmps);
    if (!sat) {
      delete Aux_slvr;
      return(true);
    }
    
    MvecLits Assmps1;
    bool yes = removable_bnd_pnt(Assmps1,Aux_slvr);
    if (yes) {
      rem_bnd_pnts++;
      CLAUSE D;
      gen_impl_cls(D,Assmps1);
      accept_new_clause(Slvr,D);
      accept_new_clause(Aux_slvr,D);
      Sol.push_back(D);
      new_ycubes++;
      if (max_new_ycubes > 0)
        if (new_ycubes >= max_new_ycubes)
          return(false);
      
      if (D.size() == 0) return(true);
      continue;
    }
    unrem_bnd_pnts++;
    CLAUSE B;
    blocking_clause(B);
    if (B.size() == 0) {
      delete Aux_slvr;
      return(true);
    }

    add_sat_xlit(B,C);
    accept_new_clause(Aux_slvr,B);
  }

} /* prove_red */


//
//  FORM_BLOCKING_CLAUSE
//
void EgPqe::form_blocking_clause(CUBE &B,SCUBE &Ya)
{

  SCUBE::iterator pnt;

  for (pnt = Ya.begin(); pnt != Ya.end(); pnt++)
    B.push_back(-(*pnt));


} /* form_blocking_clause */

//
// MANDATORY
//
bool EgPqe::mandatory(CUBE &A,int assgn,SCUBE &Ya,int cls_ind)
{

  CLAUSE &C = F[cls_ind];

  
  for (uint i=0; i < C.size(); i++) {
    if (C[i] == assgn) continue;
    int var_ind = abs(C[i])-1;
    if (A[var_ind] == -C[i]) continue; // skip falsif. assgn
    if (Xvar[var_ind]) return(false); // sat by an x-assgn
    if (Ya.find(A[var_ind]) != Ya.end())
      return(false); // sat by an y-assgn of Ya
  }

  return(true);
} /* mandatory */

//
//  REDUCE_Y_ASSGNS
//
void EgPqe::reduce_y_assgns(SCUBE &Ya,CUBE &A)
{

  for (uint i=0; i < A.size(); i++) {
    if (Xvar[i]) continue;
    CUBE *pClauses;
    if (A[i] < 0) pClauses = &Flits0[i];
    else pClauses = &Flits1[i];

    bool crucial = false;
    for (uint j=0; j < pClauses->size(); j++) {
      int cls_ind = (*pClauses)[j];
      if (mandatory(A,A[i],Ya,cls_ind)) {
        crucial = true;
        break;
      }
    }

    if (crucial) Ya.insert(A[i]);
  }

} /* reduce_y_assgns */

//
// BLOCKING_CLAUSE
//
void EgPqe::blocking_clause(CLAUSE &B)
{

 CUBE A;

 form_sat_assgn(A,Slvr);
 assert((int) A.size() == num_vars);
 SCUBE Ya;

 reduce_y_assgns(Ya,A);
 form_blocking_clause(B,Ya);

} /* blocking_clause */

//
//  REMOVABLE_BND_PNT
//
bool EgPqe::removable_bnd_pnt(MvecLits &Assmps,Msolver *Aux_slvr)
{

  CUBE Ya;

  extr_y_assgns(Ya,Aux_slvr);
  
  add_assumps(Assmps,Ya);
  bool sat = Slvr->solve(Assmps);
  return(!sat); // if (!sat) the boundary point is removable

} /* removable_bnd_pnt */


//
//  EXTR_Y_ASSGNS
//
void EgPqe::extr_y_assgns(CUBE &Ya,Msolver *Aux_slvr)
{
  MboolVec &S = Aux_slvr->model;
  for (int i=0; i < num_vars; i++) {
    if (Xvar[i]) continue;
    if (S[i] == Mtrue) Ya.push_back(i+1);
    else Ya.push_back(-(i+1));
  }
} /* extr_y_assgns */


//
//  FORM_SAT_ASSGN
//
void EgPqe::form_sat_assgn(CUBE &A,Msolver *Solver)
{
  MboolVec &S = Solver->model;
  for (int i=0; i < num_vars; i++) 
    if (S[i] == Mtrue) A.push_back(i+1);
    else A.push_back(-(i+1));
} /* form_sat_assgn */
